package org.ibex.arenaj;
import soot.*;
import soot.jimple.*;
import soot.util.*;
import java.io.*;
import java.util.*;

public class Main extends SceneTransformer {    

    public static final int initialSize = 100;

    private static Main instance = new Main();
    private Main() { }
    public static Main v() { return instance; }
   
    public static void main(String[] args) throws Exception {
        if(args.length == 0) {
            System.out.println("Syntax: java " + v().getClass().getName() + " [soot options]");
            System.exit(0);
        }
        PackManager.v().getPack("wjtp").add(new Transform("wjtp.tx", Main.v()));
        String[] args2 = new String[args.length + 14];
        System.arraycopy(args, 0, args2, 0, args.length-1);
        String sootcp =
            System.getProperty("java.class.path") + 
            File.pathSeparator +
            System.getProperty("sun.boot.class.path");
        String extDirs = System.getProperty("java.ext.dirs");
        if (extDirs != null) {
            StringTokenizer st = new StringTokenizer(extDirs, File.pathSeparatorChar+"");
            while(st.hasMoreTokens()) {
                String goo = st.nextToken();
                System.out.println("goo " + goo);
                File f = new File(goo);
                if (!f.exists() || !f.isDirectory()) continue;
                String[] jars = f.list();
                for(int i=0; i<jars.length; i++)
                    sootcp += File.pathSeparator + f.getPath() + File.separatorChar + jars[i];
            }
        }
        args2[args.length - 1] = "-w";
        args2[args.length + 0] = "-p";
        args2[args.length + 1] = "cg";
        args2[args.length + 2] = "enabled:false";
        args2[args.length + 3] = "-p";
        args2[args.length + 4] = "wjap";
        args2[args.length + 5] = "enabled:false";
        args2[args.length + 6] = "-p";
        args2[args.length + 7] = "jtp";
        args2[args.length + 8] = "enabled:false";
        //args2[args.length + 9] = "-keep-line-number";
        args2[args.length + 9] = "-allow-phantom-refs";
        args2[args.length + 10] = "-allow-phantom-refs";
        args2[args.length + 11] = "-cp";
        args2[args.length + 12] = sootcp;
        System.out.println("sootcp => " + sootcp);
        args2[args.length + 13] = args[args.length-1];
        for(int i=0; i<args2.length; i++) System.out.println("args2["+i+"] = " + args2[i]);
        soot.Main.main(args2);
    }

    static int tfr = 0;
    public static Local viaLocal(Value v, Body b) {
        Local l = Jimple.v().newLocal("tmpRef" + (tfr++), v.getType());
        b.getLocals().add(l);
        if (v instanceof IdentityRef) {
            b.getUnits().addFirst(Jimple.v().newIdentityStmt(l, v));
        } else {
            b.getUnits().addFirst(Jimple.v().newAssignStmt(l, v));
        }
        return l;
    }
    public static Local viaLocal(Value v, Body b, Unit u) {
        Local l = Jimple.v().newLocal("tmpRef" + (tfr++), v.getType());
        b.getLocals().add(l);
        if (v instanceof IdentityRef) {
            b.getUnits().insertBefore(Jimple.v().newIdentityStmt(l, v), u);
        } else {
            b.getUnits().insertBefore(Jimple.v().newAssignStmt(l, v), u);
        }
        return l;
    }

    HashMap map = new HashMap();
    HashSet size_added = new HashSet();

    public boolean isGladiator(SootClass c) { return c.implementsInterface("org.ibex.arenaj.Gladiator"); }
    public boolean isGladiatorField(SootField f)
        { return isGladiator(f.getDeclaringClass()) && !f.getName().equals("this") && f.getName().indexOf('$')==-1; }
    public boolean isGladiatorFieldRef(SootFieldRef f) {
        return isGladiator(f.declaringClass()) && !f.name().equals("this") && f.name().indexOf('$')==-1;
    }
    public SootField getGladiatorField(SootField f) { return getGladiatorField(f.makeRef()); }
    public SootField getGladiatorField(SootFieldRef f) {
        SootClass c  = f.declaringClass();
        SootClass oc = Scene.v().getSootClass(c.getName().substring(0, c.getName().lastIndexOf('$')));
        String sig = f.declaringClass().getName()+"."+f.name();
        if (map.get(sig) != null) return (SootField)map.get(sig);
        Type t = f.type();
        if (t instanceof RefType && isGladiator(((RefType)t).getSootClass())) t = IntType.v();
        SootField nf = new SootField(c.getShortName().substring(c.getShortName().lastIndexOf('$')+1) + "$" + f.name(),
                                     t.makeArrayType(),
                                     0);
        oc.addField(nf);

        Body body = getInitBody(oc);
        Expr newArr = Jimple.v().newNewArrayExpr(t, IntConstant.v(initialSize));
        Local newArrLocal = Jimple.v().newLocal("tmpRef" + (tfr++), f.type().makeArrayType());
        body.getLocals().add(newArrLocal);
        InstanceFieldRef ifr = Jimple.v().newInstanceFieldRef(thisLocal(c,body), nf.makeRef());
        body.getUnits().addFirst(Jimple.v().newAssignStmt(ifr, newArrLocal));
        body.getUnits().addFirst(Jimple.v().newAssignStmt(newArrLocal, newArr));

        map.put(sig, nf);
        return nf;
    }

    public Body getInitBody(SootClass c) {
        c.setApplicationClass();
        List list = new LinkedList();
        SootMethod m = c.getMethod("<init>", list);
        if (!m.hasActiveBody()) {
            JimpleBody b = (JimpleBody)Jimple.v().newBody(m);
            m.setActiveBody(b);
        }
        return m.getActiveBody();
    }

    public Local thisLocal(SootClass c, Body b) { return b.getThisLocal(); }
    public SootFieldRef getGladiatorFieldSizeRef(SootClass c) {
        SootClass mc = getParent(c);
        String name = c.getShortName().substring(c.getShortName().lastIndexOf('$')+1) + "$$size";
        if (map.get(name) == null) {
            SootField f = new SootField(name, IntType.v());
            mc.addField(f);

            Body body = getInitBody(mc);
            InstanceFieldRef ifr = Jimple.v().newInstanceFieldRef(thisLocal(c,body), f.makeRef());
            body.getUnits().addFirst(Jimple.v().newAssignStmt(ifr, IntConstant.v(0)));

            map.put(c, f);
        }
        return Scene.v().makeFieldRef(mc, name, IntType.v(), false);
    }
    public SootFieldRef getGladiatorFieldMaxRef(SootClass c) {
        SootClass mc = getParent(c);
        String name = c.getShortName().substring(c.getShortName().lastIndexOf('$')+1) + "$$max";
        if (map.get(name) == null) {
            SootField f = new SootField(name, IntType.v());
            mc.addField(f);

            Body body = getInitBody(mc);
            InstanceFieldRef ifr = Jimple.v().newInstanceFieldRef(thisLocal(c,body), f.makeRef());
            body.getUnits().addFirst(Jimple.v().newAssignStmt(ifr, IntConstant.v(initialSize)));

            map.put(c, f);
        }
        return Scene.v().makeFieldRef(mc, name, IntType.v(), false);
    }

    public boolean isGladiatorType(Type t) { return (t instanceof RefType) && isGladiator(((RefType)t).getSootClass()); }


    

    boolean done = false;
    public void internalTransform(String phaseName, Map options) {
        System.out.println("begun");
        List ac = new LinkedList();
        for(Iterator it = Scene.v().getClasses().iterator(); it.hasNext();) {
            SootClass sc = (SootClass)it.next();
            sc.setLibraryClass();
        }
        for(Iterator it = Scene.v().getClasses().iterator(); it.hasNext();) {
            SootClass sc = (SootClass)it.next();
            if (isGladiator(sc)) {
                System.out.println("  application class: " + sc.getName());
                System.out.println("  application class: " + getParent(sc).getName());
                sc.setApplicationClass();
                getParent(sc).setApplicationClass();
                ac.add(sc);
                ac.add(getParent(sc));
                for(Iterator i2 = sc.getMethods().iterator(); i2.hasNext();) {
                    SootMethod m = (SootMethod)i2.next();
                    if (m.isConcrete()) m.retrieveActiveBody();
                }
                for(Iterator i2 = getParent(sc).getMethods().iterator(); i2.hasNext();) {
                    SootMethod m = (SootMethod)i2.next();
                    if (m.isConcrete()) m.retrieveActiveBody();
                }
            }
        }
        for(Iterator it = ac.iterator(); it.hasNext();) {
            SootClass sc = (SootClass)it.next();
            if (sc.isInterface()) continue;
            if (!isGladiator(sc)) continue;
            System.out.println("fixing Gladiator class: " + sc);
            fixClass(sc);
        }
        for(Iterator it = ac.iterator(); it.hasNext();) {
            SootClass sc = (SootClass)it.next();
            if (sc.isInterface()) continue;
            System.out.println("        updating class: " + sc);
            nuke(sc);
        }
    }

    public SootClass getParent(SootClass sc) {
        return Scene.v().getSootClass(sc.getName().substring(0, sc.getName().lastIndexOf('$')));
    }

    public void fixClass(SootClass sc) {
        SootClass mc = getParent(sc);
        String incFuncName = sc.getShortName().substring(sc.getShortName().lastIndexOf('$')+1) + "$$inc";
        SootMethod method = new SootMethod(incFuncName, new LinkedList(),
                                           IntType.v(), Modifier.PRIVATE,
                                           new LinkedList());
        mc.addMethod(method);
        Body body = Jimple.v().newBody(method);
        method.setActiveBody(body);
        ((JimpleBody)body).insertIdentityStmts();
        Local l = Jimple.v().newLocal("tmpRef" + (tfr++), IntType.v());
        body.getLocals().add(l);
        Local l2 = Jimple.v().newLocal("tmpRef" + (tfr++), IntType.v());
        body.getLocals().add(l2);
        Local l3 = Jimple.v().newLocal("tmpRef" + (tfr++), IntType.v());
        body.getLocals().add(l3);
        InstanceFieldRef sfr = Jimple.v().newInstanceFieldRef(body.getThisLocal(), getGladiatorFieldSizeRef(sc));
        body.getUnits().add(Jimple.v().newAssignStmt(l, sfr));
        body.getUnits().add(Jimple.v().newAssignStmt(l2, Jimple.v().newAddExpr(l, IntConstant.v(1))));
        InstanceFieldRef maxField = Jimple.v().newInstanceFieldRef(body.getThisLocal(), getGladiatorFieldMaxRef(sc));
        body.getUnits().add(Jimple.v().newAssignStmt(l3, maxField));
        Stmt stmt = Jimple.v().newReturnStmt(l2);
        body.getUnits().add(Jimple.v().newIfStmt(Jimple.v().newLtExpr(l2, l3), stmt));

        Local l4 = Jimple.v().newLocal("tmpRef" + (tfr++), IntType.v());
        body.getLocals().add(l4);
        body.getUnits().add(Jimple.v().newAssignStmt(l4, Jimple.v().newShlExpr(l3, IntConstant.v(1))));
        body.getUnits().add(Jimple.v().newAssignStmt(maxField, l4));

        for(Iterator it = sc.getFields().iterator(); it.hasNext();) {
            SootField f = getGladiatorField((SootField)it.next());
            InstanceFieldRef ifr = Jimple.v().newInstanceFieldRef(body.getThisLocal(), f.makeRef());
            Local ll0 = Jimple.v().newLocal("tmpRef" + (tfr++), f.getType());
            body.getLocals().add(ll0);
            body.getUnits().add(Jimple.v().newAssignStmt(ll0, ifr));
            Local ll = Jimple.v().newLocal("tmpRef" + (tfr++), f.getType());
            body.getLocals().add(ll);
            body.getUnits().add(Jimple.v().newAssignStmt(ll,
                                                         Jimple.v().newNewArrayExpr(((ArrayType)f.getType()).getElementType(),
                                                                                    l4)));
            Type ot = Scene.v().getSootClass("java.lang.Object").getType();
            List types = new LinkedList();
            types.add(ot);
            types.add(IntType.v());
            types.add(ot);
            types.add(IntType.v());
            types.add(IntType.v());
            SootMethodRef arrayCopy =
                Scene.v().makeMethodRef(Scene.v().getSootClass("java.lang.System"),
                                        "arraycopy",
                                        types,
                                        VoidType.v(),
                                        true);
            List args = new LinkedList();
            args.add(ll0);
            args.add(IntConstant.v(0));
            args.add(ll);
            args.add(IntConstant.v(0));
            args.add(l3);
            body.getUnits().add(Jimple.v().newInvokeStmt(Jimple.v().newStaticInvokeExpr(arrayCopy, args)));
            body.getUnits().add(Jimple.v().newAssignStmt(ifr,ll));
        }
        for(Iterator it = sc.getMethods().iterator(); it.hasNext();) {
            SootMethod m = (SootMethod)it.next();
            if (!m.isConcrete()) continue;
            boolean doremove = true;
            Body mbody = m.getActiveBody();
            if (isGladiator(m.getDeclaringClass()) && m.getName().equals("<init>")) {
                doremove = false;
                SootClass c = m.getDeclaringClass();
                String name = /*c.getShortName().substring(c.getShortName().lastIndexOf('$')+1) + */"$init";
                List li = m.getParameterTypes();
                //li.add(IntType.v());
                c.removeMethod(m);
                SootMethod nm = new SootMethod(name, li, isGladiatorType(m.getReturnType()) ? IntType.v() : m.getReturnType());
                JimpleBody bod = Jimple.v().newBody(nm);
                bod.importBodyContentsFrom(m.getActiveBody());
                nm.setActiveBody(bod);
                m = nm;
                mbody = bod;
                mbody.getUnits().remove(((JimpleBody)mbody).getFirstNonIdentityStmt());
                mbody.getUnits().remove(((JimpleBody)mbody).getFirstNonIdentityStmt());
            } else {
                System.out.println("examining " + sc.getName() + "." + m.getSignature());
                m.retrieveActiveBody();
            }
            if (m.isStatic()) continue;

            String name = sc.getShortName().substring(sc.getShortName().lastIndexOf('$')+1) + "$$" + m.getName();
            List list = new LinkedList();
            list.addAll(m.getParameterTypes());
            list.add(IntType.v());
            for(Iterator i = list.iterator(); i.hasNext();) System.out.println(i.next());
            SootMethod m2 = new SootMethod(name, list, m.getReturnType());
            getParent(sc).addMethod(m2);

            JimpleBody ab = (JimpleBody)Jimple.v().newBody(m2);
            ab.importBodyContentsFrom(mbody);
            m2.setActiveBody(ab);
            Local loc = Jimple.v().newLocal("tmpRef" + (tfr++), getParent(sc).getType());
            ab.getLocals().add(loc);
            // FIXME: insert assignment to this

            for(Iterator z = ab.getLocals().iterator(); z.hasNext();) {
                Local loc2 = (Local)z.next();
                if (isGladiatorType(loc2.getType())) {
                    loc2.setType(IntType.v());
                }
            }

            Chain units = ab.getUnits();
            boolean touched = false;
            Local loc0 = Jimple.v().newLocal("tmpRef" + (tfr++), getParent(sc).getType());
            ab.getLocals().add(loc0);
            for(Iterator stmtIt = units.snapshotIterator(); stmtIt.hasNext();) {
                Stmt s = (Stmt) stmtIt.next();
                if (s instanceof IdentityStmt) {
                    IdentityStmt is = (IdentityStmt)s;
                    Local left = (Local)is.getLeftOp();
                    if (is.getRightOp() instanceof ThisRef) {
                        left.setType(IntType.v());
                        is.getRightOpBox().setValue(Jimple.v().newParameterRef(IntType.v(), m.getParameterCount()));
                        if (!touched) {
                            units.addFirst(Jimple.v().newIdentityStmt(loc0, Jimple.v().newThisRef(getParent(sc).getType())));
                            touched = true;
                        }
                    }
                }

                for(Iterator i = s.getUseAndDefBoxes().iterator(); i.hasNext();) {
                    Object o = i.next();
                    if (o instanceof ValueBox) {
                        ValueBox vb = (ValueBox)o;
                        o = vb.getValue();
                        /*
                        if (o instanceof Local && isGladiatorType(((Local)o).getType())) {
                            System.out.println("thunking");
                            vb.setValue(loc0);
                        }
                        */
                        if (vb.getValue() instanceof ThisRef) {
                            System.out.println(s);
                            vb.setValue(loc);
                        }
                    }
                }

            }
            
            if (doremove) sc.removeMethod(m);

        }
        body.getUnits().add(stmt);
    }

    public void nuke(SootClass c) {
        for(Iterator it = c.getFields().iterator(); it.hasNext();) {
            SootField f = (SootField)it.next();
            Type t = f.getType();
            if (t instanceof RefType) {
                RefType rt = (RefType)t;
                SootClass sc = rt.getSootClass();
                if (isGladiator(sc)) f.setType(IntType.v());
            } else if (t instanceof ArrayType) {
                ArrayType at = (ArrayType)t;
                t = at.getElementType();
                if (!(t instanceof RefType)) continue;                
                RefType rt = (RefType)t;
                SootClass sc = rt.getSootClass();
                if (isGladiator(sc)) f.setType(IntType.v().makeArrayType());
            }
        }

        List list = new LinkedList(); list.addAll(c.getMethods());
        for(Iterator it = list.iterator(); it.hasNext();) {
            SootMethod m = (SootMethod)it.next();
            Body b = null;
            if (m.hasActiveBody()) b = fixBody(m.getActiveBody(), c, m);
            List l2 = new LinkedList();
            List l = m.getParameterTypes();
            for(Iterator it2 = l.iterator(); it2.hasNext();) {
                Type t = (Type)it2.next();
                l2.add(isGladiatorType(t) ? IntType.v() : t);
            }
            Type t = m.getReturnType();
            if (isGladiatorType(t)) {
                t = IntType.v();
                if (m.hasActiveBody()) {
                    Body bod = m.getActiveBody();
                    for(Iterator stmtIt = bod.getUnits().snapshotIterator(); stmtIt.hasNext();) {
                        Stmt s = (Stmt) stmtIt.next();
                        if (s instanceof ReturnStmt) {
                            if (((ReturnStmt)s).getOp().getType() instanceof NullType) {
                                ((ReturnStmt)s).getOpBox().setValue(IntConstant.v(-1));
                            }
                        }
                    }
                }
            }
            String name = m.getName();
            SootMethod meth = new SootMethod(name, l2, isGladiatorType(t) ? IntType.v() : t, m.getModifiers());
            if (b != null) {
                JimpleBody b2 = Jimple.v().newBody(meth);
                b2.importBodyContentsFrom(b);
                meth.setActiveBody(b2);
            }
            c.removeMethod(m);
            c.addMethod(meth);
        }

    }

    protected Body fixBody(Body body, SootClass ownerClass, SootMethod smeth) {
        if (body instanceof JimpleBody) {
            JimpleBody b2 = Jimple.v().newBody(smeth);
            b2.importBodyContentsFrom(body);
            body = b2;
        }
        Chain units = body.getUnits();
        System.out.println("fixBody("+body.getMethod()+")");
        for(Iterator it = body.getLocals().snapshotIterator(); it.hasNext();) {
            Local l = (Local)it.next();
            if (isGladiatorType(l.getType())) l.setType(IntType.v());
        }
        for(int qq=0; qq<2; qq++) for(Iterator stmtIt = units.snapshotIterator(); stmtIt.hasNext();) {
            Stmt s = (Stmt) stmtIt.next();
            if (s instanceof DefinitionStmt) {
                DefinitionStmt ds = (DefinitionStmt)s;
                if (ds.getLeftOp().getType() instanceof PrimType && ds.getRightOp().getType() instanceof NullType) {
                    ds.getRightOpBox().setValue(IntConstant.v(-1));
                }
            }
            if (isGladiatorType(smeth.getReturnType()) && s instanceof ReturnStmt) {
                if (((ReturnStmt)s).getOp().getType() instanceof NullType) {
                    ((ReturnStmt)s).getOpBox().setValue(IntConstant.v(-1));
                }
            }
            List l = s.getUseAndDefBoxes();
            for(Iterator it = l.iterator(); it.hasNext();) {
                Object o = it.next();
                if (o instanceof ValueBox) {
                    ValueBox vb = (ValueBox)o;
                    Value v = vb.getValue();
                    
                    if (v instanceof BinopExpr) {
                        BinopExpr boe = (BinopExpr)v;
                        if (boe.getOp1().getType() instanceof PrimType && boe.getOp2().getType() instanceof NullType) {
                            boe.setOp2(IntConstant.v(-1));
                        }
                        if (boe.getOp2().getType() instanceof PrimType && boe.getOp1().getType() instanceof NullType) {
                            boe.setOp1(IntConstant.v(-1));
                        }
                    }

                    if (v instanceof NewExpr) {
                        NewExpr ne = (NewExpr)v;
                        if (isGladiatorType(ne.getBaseType())) {
                            SootClass sc = ((RefType)ne.getBaseType()).getSootClass();
                            SootClass mc = getParent(sc);
                            String incFuncName = sc.getShortName().substring(sc.getShortName().lastIndexOf('$')+1) + "$$inc";
                            SootMethodRef smr = Scene.v().makeMethodRef(mc, incFuncName, new LinkedList(), IntType.v(), false);
                            Expr invokeExpr = Jimple.v().newSpecialInvokeExpr(thisLocal(mc,body), smr);
                            Local ll = viaLocal(invokeExpr, body, s);
                            vb.setValue(ll);
                            v = ll;
                            continue;
                        } 
                    } else

                    if (v instanceof InvokeExpr) {
                        InvokeExpr ie = (InvokeExpr)v;
                        SootMethodRef mr = ie.getMethodRef();
                        String name = mr.name();
                        if (v instanceof InstanceInvokeExpr && isGladiator(mr.declaringClass())) {
                            InstanceInvokeExpr iie = (InstanceInvokeExpr)v;
                            List li = new LinkedList();
                            li.addAll(iie.getArgs());
                            LinkedList pl = new LinkedList();
                            for(Iterator it2 = mr.parameterTypes().iterator(); it2.hasNext();) {
                                Type t = (Type)it2.next();
                                pl.add(isGladiatorType(t) ? IntType.v() : t);
                            }
                            if (mr.name().equals("<init>") && isGladiator(mr.declaringClass())) {
                                name = "$init";
                                //li.remove(0);
                                //pl.remove(0);
                            }
                            pl.add(IntType.v());
                            li.add(viaLocal(iie.getBase(),body,s));
                            SootClass sc = mr.declaringClass();
                            name = sc.getShortName().substring(sc.getShortName().lastIndexOf('$')+1) + "$$" + name;
                            mr = Scene.v().makeMethodRef(getParent(sc),
                                                         name,
                                                         pl,
                                                         isGladiatorType(mr.returnType()) ? IntType.v() : mr.returnType(),
                                                         false);
                            ie = Jimple.v().newVirtualInvokeExpr(body.getThisLocal(), mr, li);
                            vb.setValue(v = ie);

                        } else if (!(v instanceof StaticInvokeExpr)) {
                            List l0 = mr.parameterTypes();
                            List l2 = new LinkedList();
                            for(Iterator it2 = l0.iterator(); it2.hasNext();) {
                                Type t = (Type)it2.next();
                                l2.add(isGladiatorType(t) ? IntType.v() : t);
                            }
                            mr = Scene.v().makeMethodRef(mr.declaringClass(),
                                                         mr.name(),
                                                         l2,
                                                         isGladiatorType(mr.returnType()) ? IntType.v() : mr.returnType(),
                                                         mr.isStatic());
                            ie.setMethodRef(mr);
                            vb.setValue(v = ie);                            
                        }

                        for(int i=0; i<ie.getArgCount(); i++) {
                            ValueBox b = ie.getArgBox(i);
                            Value val = b.getValue();
                            if (mr.parameterType(i) instanceof RefType && val.getType() instanceof PrimType) {
                                SootClass intClass = Scene.v().getSootClass("java.lang.Integer");
                                List typelist = new LinkedList();
                                typelist.add(IntType.v());
                                SootMethod intMethod = intClass.getMethod("<init>", typelist);
                                Local loc = viaLocal(Jimple.v().newNewExpr(RefType.v(intClass)), body, s);
                                List list = new LinkedList();
                                list.add(val);
                                units.insertBefore(Jimple.v().newInvokeStmt(Jimple.v().newSpecialInvokeExpr(loc,
                                                                                                            intMethod.makeRef(),
                                                                                                            list)),
                                                   s);
                                b.setValue(loc);
                            }
                            if (val != null && val.getType() instanceof NullType && mr.parameterType(i) instanceof IntType) {
                                b.setValue(IntConstant.v(-1));
                            }
                        }


                    } else if (v instanceof CastExpr) {
                        CastExpr ce = (CastExpr)v;
                        if (isGladiatorType(ce.getCastType())) {
                            SootClass mc = getParent(((RefType)ce.getCastType()).getSootClass());
                            SootClass ic = Scene.v().getSootClass("java.lang.Integer");
                            ce.setCastType(ic.getType());
                            // FIXME deal with null
                            SootMethodRef mr = Scene.v().makeMethodRef(ic, "intValue", new LinkedList(), IntType.v(), false);
                            InvokeExpr iie = Jimple.v().newVirtualInvokeExpr(viaLocal(ce, body, s),
                                                                             mr,
                                                                             new LinkedList());
                            vb.setValue(viaLocal(iie, body, s));
                            qq = 0;
                            break;
                        }

                    } else if (v instanceof InstanceFieldRef) {
                        InstanceFieldRef ifr = (InstanceFieldRef)v;
                        SootFieldRef fr = ifr.getFieldRef();
                        Type t = fr.type();
                        if (isGladiator(fr.declaringClass()) && fr.name().equals("this$0")) {
                            System.out.println("******************");
                            vb.setValue(body.getThisLocal());
                        } else if (isGladiatorFieldRef(fr)) {
                            SootClass mc = getParent(fr.declaringClass());
                            SootFieldRef sf = getGladiatorField(fr).makeRef();
                            InstanceFieldRef sfr = Jimple.v().newInstanceFieldRef(thisLocal(mc, body), sf);
                            System.out.println("s is " + s);
                            ArrayRef ar = Jimple.v().newArrayRef(viaLocal(sfr, body, s), ifr.getBase());
                            vb.setValue(ar);
                        }
                        if ((t instanceof RefType) && isGladiator(((RefType)t).getSootClass())) {
                            SootClass tc = ((RefType)t).getSootClass();
                            SootClass mc = getParent(tc);
                            ifr.setFieldRef(Scene.v().makeFieldRef(mc, fr.name(), IntType.v(), false));
                        }
                    }

                }
            }
        }
        body.validate();
        return body;
    }
}
